/* Copyright (C) 2010 Philippe Lindheimer (philippe@freepbx.org)
 */

/* CCSS Asterisk Device States Proposal

Proposal to Add Asterisk Device State information and callbacks to the
Call Completion Supplemental Services for generic agents. Note: this is not
100% a c source file, I just named it that way to get syntax highlighting
to make this very readable since there is a lot of code.

There are currently not many devices that have native support for CCSS and
even as they come available there may be other reasons why one may choose not
to take advantage of the native abilities and stick with the generic
implementation which is quite capable and could be greatly enhanced by adding
device state capabilities that a phone could subscribe to with a BLF key in
conjuction with Asterisk hints.

The advantages of this would allow a single button that could be programmed to
trigger a CCSS related feature code that can be used to both Request and
Cancel CCSS requests as well as display the current state of a CCSS Request.

By example, I may have a single button that when no lit, there is no active 
CCSS request. When I press that button, my dialplan can query the state
DEVICE_STATE() associated with that Caller to determine whether they should be
calling CallCompletionRequest() or CallCompletionCancel(). If there is
currently a pending request, then the dialplan would Cancel() it. This also has
the advantage of showing the true state of a Request() which is an
asynchronous call and even when Request() thinks it was successful, the actual
request could ultimately fail. Once lit, further feedback can now be provided
to the Caller about the current state of their Request() since it will be
updated by the CCSS State Machine as appropriate.

The DEVICE_STATE mappings should be configurable since the BLF being 
used on a give phone type may vary. The idea is to allow some level of 
customization as to the phone's behavior.

As an example, I may want my BLF key to go solid once I have requested a
callback. I may then want my led to blink (typically ringing) when either the
callback is in process which is a visual indication that the incoming call is
the desired callback, or I may want it to blink when the callee is ready buy I
am busy, giving me a visual indication that the target is available as I may
want to get off the line so that the callback can be successful.

Proposed Default Values:

cc_available_devstate        = AST_DEVICE_NOT_INUSE
cc_caller_offered_devstate 	 = AST_DEVICE_NOT_INUSE
cc_caller_requested_devstate = AST_DEVICE_NOT_INUSE
cc_active_devstate           = AST_DEVICE_NOT_INUSE
cc_callee_ready_devstate     = AST_DEVICE_RINGING
cc_caller_busy_devstate      = AST_DEVICE_RINGINUSE
cc_recalling_devstate        = AST_DEVICE_RINGING
cc_complete_devstate         = AST_DEVICE_NOT_INUSE
cc_failed_devstate           = AST_DEVICE_NOT_INUSE

The proposal is to send device state information back via the
ast_devstate_prov_add callback for any device (or at least generic device)
as they traverse through the state machine. We simply provide a map
between CC_STATE values and corresponding AST_DEVICE state values.

We could then generate hints against these States similar to what is possible
today with Custom Devstates or MeetMe states. By example, I may have an extension 
3000 that is currently associated with device SIP/3000. I could then create a 
feature code for that extension that may look something like:

exten => *823000,hint,ccss:sip/3000

One would then subscribe a blf button to *823000 which would point to the dialplan 
that handled my CCSS Requests/Cancels using the available DEVICE_STATE() 
information about ccss:sip/3000 to make the decsions what to do.

Below are the required callbacks, functions, structures and patch to ccss.c
to conceptually implement this proposed device state functionality.


SHORT SUMMARY OF IMPLENTATION:

The crux of the proposal takes advantage of the central function:

  cc_do_state_change()

where all state changes occur. We simple add a call upon a successful
state change of a generic device to:

	ccss_notify_device_state_change()

which updates the state information using ast_devstate_changed(). We also
provide the required callback function:

  ccss_device_state()

initialized with ast_devstate_prov_add() which looks up and returns the
current state for the supplied device.

initialize_cc_devstate_map() is provided to change the default cc_state
to devstate mapping in ccss.conf.
*/

/* default values mapping from cc_state to ast_dev_state */

#define CC_AVAILABLE_DEVSTATE_DEFAULT        AST_DEVICE_NOT_INUSE
#define CC_CALLER_OFFERED_DEVSTATE_DEFAULT   AST_DEVICE_NOT_INUSE
#define CC_CALLER_REQUESTED_DEVSTATE_DEFAULT AST_DEVICE_NOT_INUSE
#define CC_ACTIVE_DEVSTATE_DEFAULT           AST_DEVICE_INUSE
#define CC_CALLEE_READY_DEVSTATE_DEFAULT     AST_DEVICE_RINGING
#define CC_CALLER_BUSY_DEVSTATE_DEFAULT      AST_DEVICE_RINGINUSE
#define CC_RECALLING_DEVSTATE_DEFAULT        AST_DEVICE_RINGING
#define CC_COMPLETE_DEVSTATE_DEFAULT         AST_DEVICE_NOT_INUSE
#define CC_FAILED_DEVSTATE_DEFAULT           AST_DEVICE_NOT_INUSE

/* Structure to map cc_state to an ast_device_state. Initialized here but
 * configured through ccss.conf in iniitalize_cc_devstate_map() function. 
 */
struct cc_state_2_devstate {
	enum cc_state state;
	enum ast_device_state devstate;
};

/* these are the defaults, ccss.conf can set these differently */
static struct cc_state_2_devstate cc_state_to_devstate_map[] = {
	{CC_AVAILABLE,          CC_AVAILABLE_DEVSTATE_DEFAULT},
	{CC_CALLER_OFFERED,     CC_CALLER_OFFERED_DEVSTATE_DEFAULT},
	{CC_CALLER_REQUESTED,   CC_CALLER_REQUESTED_DEVSTATE_DEFAULT},
	{CC_ACTIVE,             CC_ACTIVE_DEVSTATE_DEFAULT},
	{CC_CALLEE_READY,       CC_CALLEE_READY_DEVSTATE_DEFAULT},
	{CC_CALLER_BUSY,        CC_CALLER_BUSY_DEVSTATE_DEFAULT},
	{CC_RECALLING,          CC_RECALLING_DEVSTATE_DEFAULT},
	{CC_COMPLETE,           C_COMPLETE_DEVSTATE_DEFAULT
	{CC_FAILED,             CC_FAILED_DEVSTATE_DEFAULT},
};

/*! \brief Callback for devicestate providers */
static enum ast_device_state ccss_device_state(const char *device_name)
{

	unsigned long match_flags;
	int res;

	match_flags = MATCH_NO_REQUEST;
	if (!(core_instance = ao2_t_callback_data(cc_core_instances, 0, match_agent, device_name, &match_flags, "Find Core Instance for ccss_device_state reqeust."))) {
		ast_debug(4, "Couldn't find a core instance for caller %s\n", device_name);
		return cc_state_to_devstate(CC_FAILED);
	}

	ast_debug(4, "Core %d: Found core_instance for caller %s in state %s\n",
			core_instance->core_id, device_name, cc_state_to_string(core_instance->current_state));

	/* TODO: at this point do I care if it is generic or not, it's still valid to return the state right?
	 *       I would still have to make sure to call cc_unref()
	 */
	if (strcmp(core_instance->agent->callbacks->type, "generic")) {
		ast_debug(4, "Core %d: Device State is only for generic agent types.\n",
				core_instance->core_id);
		cc_unref(core_instance, "Unref core_instance since ccss_device_state was called with native agent");
		return cc_state_to_devstate(CC_FAILED);
	}
	return cc_state_to_devstate(state);
}

/*! \brief Notify Device State Changes from CC STATE MACHINE */
static void ccss_notify_device_state_change(const char *device, enum cc_state state)
{
  enum ast_device_state devstate;

	devstate = cc_state_to_devstate(state);

	ast_debug(4, "Notification of CCSS state change to '%s', device state '%s' for device '%s'", 
		cc_state_to_string(state), ast_devstate2str(devstate)), device;

	ast_devstate_changed(devstate, "ccss:%s", device);
}

/* \brief lookup the ast_device_state mapped to cc_state */
static enum ast_device_state cc_state_to_devstate(enum cc_state state)
{
	return cc_state_to_devstate_map[state].devstate;
}

/* \brief initializes cc_state_to_devstate_map from ccss.conf */
static void initialize_cc_devstate_map(void)
{
	struct ast_config *cc_config;
	struct ast_flags config_flags = {0,};
	enum ast_device_state this_devstate;

	/* don't need to set these, they are initialized in the struct:
	 *
	cc_state_to_devstate_map[CC_AVAILABLE].devstate = CC_AVAILABLE_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_CALLER_OFFERED].devstate = CC_CALLER_OFFERED_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_CALLER_REQUESTED].devstate = CC_CALLER_REQUESTED_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_ACTIVE].devstate = CC_ACTIVE_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_CALLEE_READY].devstate = CC_CALLEE_READY_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_CALLER_BUSY].devstate = CC_CALLER_BUSY_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_RECALLING].devstate = CC_RECALLING_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_COMPLETE].devstate = CC_COMPLETE_DEVSTATE_DEFAULT;
	cc_state_to_devstate_map[CC_FAILED].devstate = CC_FAILED_DEVSTATE_DEFAULT;
	*/

	cc_config = ast_config_load2("ccss.conf", "ccss", config_flags);
	if (!cc_config || cc_config == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "Could not find valid ccss.conf file. Using cc_[state]_devstate defaults\n");
		return;
	}

	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_available_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_AVAILABLE].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_caller_offered_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_CALLER_OFFERED].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_caller_requested_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_CALLER_REQUESTED].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_active_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_ACTIVE].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_callee_ready_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_CALLEE_READY].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_caller_busy_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_CALLER_BUSY].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_recalling_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_RECALLING].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_complete_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_COMPLETE].devstate = this_devstate;
		}
	}
	if ((cc_devstate_str = ast_variable_retrieve(cc_config, "general", "cc_failed_devstate"))) {
		this_devstate = ast_devstate_val(cc_devstate_str);
		if (this_devstate != AST_DEVICE_UNKNOWN) {
			cc_state_to_devstate_map[CC_FAILED].devstate = this_devstate;
		}
	}

	ast_config_destroy(cc_config);
	return;
}

// PATCH TO ADD DEVSTATE UPDATES
//
// ---------------------------------------------------------------------------
Index: ccss.c
===================================================================
--- ccss.c	(revision 307274)
+++ ccss.c	(working copy)
@@ -2967,6 +2967,11 @@
 	core_instance->current_state = args->state;
 	res = state_change_funcs[core_instance->current_state](core_instance, args, previous_state);
 
+	/* If state change successful then notify any device state watchers of the change */
+	if (!res && !strcmp(core_instance->agent->callbacks->type, "generic")) {
+		ccss_notify_device_state_change(core_instance->device_name, core_instance->current_state);
+	}
+
 	ast_free(args);
 	cc_unref(core_instance, "Unref since state change has completed"); /* From ao2_find */
 	return res;
@@ -4220,9 +4241,14 @@
 	res |= ast_register_application2(cccancel_app, cccancel_exec, NULL, NULL, NULL);
	res |= ast_cc_monitor_register(&generic_monitor_cbs);
	res |= ast_cc_agent_register(&generic_agent_callbacks);
+
+	/* Register the device state callback for generic agents */
+	res |= ast_devstate_prov_add("ccss", ccss_device_state);
+
 	ast_cli_register_multiple(cc_cli, ARRAY_LEN(cc_cli));
 	cc_logger_level = ast_logger_register_level(CC_LOGGER_LEVEL_NAME);
 	dialed_cc_interface_counter = 1;
 	initialize_cc_max_requests();
+	initialize_cc_devstate_map();
 	return res;
 }
// ---------------------------------------------------------------------------
//
// END PATCH
